#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <arpa/inet.h>

#ifdef USE_EEL
#include <eel.h>
#else
#include <event.h>
#endif

#include "params.h"
#include "parse.h"

#define TRUE (0==0)
#define FALSE (!TRUE)

#define BUFF_SIZE 16384

#ifdef USE_EEL
eel_cb_return_value conn_dispatcher(eel_group_id gid, void *state, int fd);
#else
static void conn_dispatcher( int fd, short event, void *state );
#endif

struct dispatch_state {
#ifndef USE_EEL
    struct event ev;
#endif
    unsigned int num_conn;
};

static struct timeval tv_zero={0,0};

#ifdef USE_EEL
void init_libs(eel_group_id start_gid)
{
    struct dispatch_state *dispatch;

    eel_initialize(1);
    parsefile( options.protocol_file );

    dispatch=calloc( 1, sizeof(*dispatch) );
    dispatch->num_conn=options.num_connections;
    eel_add_timer(start_gid, conn_dispatcher, (void*)dispatch, 0);
}
#else
void init_libs()
{
    struct dispatch_state *dispatch;

    event_init();
    parsefile( options.protocol_file );

    dispatch=calloc( 1, sizeof(*dispatch) );
    dispatch->num_conn=options.num_connections;
    evtimer_set(&(dispatch->ev), conn_dispatcher, dispatch );
    evtimer_add(&(dispatch->ev), &tv_zero );
}
#endif

void new_conn();

/* Connection generating function */
#ifdef USE_EEL
eel_cb_return_value conn_dispatcher(eel_group_id gid, void *state, int fd)
{
    struct dispatch_state *dstate=(struct dispatch_state *)state;

    if( dstate->num_conn==0 ) {
        /* Finished the batch for this time */
        if( options.delay.tv_sec!=0 || options.delay.tv_usec!=0 ) {
            dstate->num_conn=options.num_connections;
            eel_add_timer(gid, conn_dispatcher, dstate, TIMEVAL_TO_MILLISECONDS(options.delay));
        }
    } else {
        new_conn();
        --dstate->num_conn;

        eel_add_timer(gid, conn_dispatcher, (void*)dstate, 0);
    }
    return EEL_SUCCESS;
}
#else
static void conn_dispatcher( int fd, short event, void *state )
{
    struct dispatch_state *dstate=(struct dispatch_state *)state;

    if( dstate->num_conn==0 ) {
        /* Finished the batch for this time */
        if( options.delay.tv_sec!=0 || options.delay.tv_usec!=0 ) {
            dstate->num_conn=options.num_connections;
            evtimer_add(&(dstate->ev), &(options.delay));
        }
    } else {
        new_conn();
        --dstate->num_conn;

        evtimer_add(&(dstate->ev), &tv_zero);
    }
}
#endif

struct conn_state {
#ifndef USE_EEL
    struct event ev;
#endif
    int fd;
    int state;
    unsigned char *buffer;
    size_t buffsize, datasize;
};

#ifdef USE_EEL
#define EV_WRITE 0x1
#define EV_READ 0x2
static void conn_handle( eel_group_id gid, int fd, short event, void *state );
eel_cb_return_value conn_handle_read(eel_group_id gid, void *state, int fd)
{
  conn_handle(gid, fd, EV_READ, state);
  return EEL_SUCCESS;
}
eel_cb_return_value conn_handle_write(eel_group_id gid, void *state, int fd)
{
  conn_handle(gid, fd, EV_WRITE, state);
  return EEL_SUCCESS;
}
static void conn_handle( eel_group_id gid, int fd, short event, void *state ) {
    struct conn_state *cstate=(struct conn_state *) state;

    if( cstate->state==num_states ) {
        eel_delete_group_id(gid);
        close( cstate->fd );
        free( cstate->buffer );
        free( state );

        return;
    }

    /* Even states mean we have to send something out */
    if( (cstate->state%2)==0 && (event&EV_WRITE)!=0 ) {
        write( cstate->fd, states[cstate->state].data, states[cstate->state].length );
        cstate->state++;
        eel_add_read(gid, conn_handle_read, state, fd);
    }
    if( (cstate->state%2)==1 && (event&EV_READ)!=0 ) {
        /* We need to receive something */
        int numread;
        unsigned char *buffer;
        const unsigned char *needle;

        cstate->buffsize=cstate->datasize+BUFF_SIZE;
        buffer=realloc( cstate->buffer, cstate->buffsize );
        if( buffer!=NULL ) {
            cstate->buffer=buffer;
        } else {
            fprintf(stderr, "Couldn't allocate buffer space\n");
            cstate->state=num_states;
            eel_combine_OR(
              eel_add_read(gid, conn_handle_read, (void*)state, fd),
              eel_add_write(gid, conn_handle_write, (void*)state, fd));

            return ;
        }
        
        cstate->datasize+=read( cstate->fd, cstate->buffer+cstate->datasize, cstate->buffsize-cstate->datasize );

        if( (needle=binbin(cstate->buffer, cstate->datasize, states[cstate->state].data,
			states[cstate->state].length ))!=NULL ) {
            cstate->datasize-=(needle-cstate->buffer)+states[cstate->state].length;
            memmove(cstate->buffer, needle+states[cstate->state].length, cstate->datasize );
            cstate->state++;
            eel_combine_OR(
              eel_add_read(gid, conn_handle_read, (void*)state, fd),
              eel_add_write(gid, conn_handle_write, (void*)state, fd));
        } else {
            /* Leave only state length bytes of the bytes read thus far in the buffer */
            memmove(cstate->buffer, cstate->buffer+cstate->datasize-states[cstate->state].length,
                    states[cstate->state].length );
            cstate->datasize=states[cstate->state].length;
            eel_add_read(gid, conn_handle_read, (void*)state, fd);
        }
    }
    
}
#else
static void conn_handle( int fd, short event, void *state ) {
    struct conn_state *cstate=(struct conn_state *) state;

    if( cstate->state==num_states ) {
        close( cstate->fd );
        free( cstate->buffer );
        free( state );

        return;
    }

    /* Even states mean we have to send something out */
    if( (cstate->state%2)==0 && (event&EV_WRITE)!=0 ) {
        write( cstate->fd, states[cstate->state].data, states[cstate->state].length );
        cstate->state++;
	event_set( &cstate->ev, cstate->fd, EV_READ, conn_handle,
		state );
	event_add( &cstate->ev, NULL );
    }
    if( (cstate->state%2)==1 && (event&EV_READ)!=0 ) {
        /* We need to receive something */
        int numread;
        unsigned char *buffer;
        const unsigned char *needle;

        cstate->buffsize=cstate->datasize+BUFF_SIZE;
        buffer=realloc( cstate->buffer, cstate->buffsize );
        if( buffer!=NULL ) {
            cstate->buffer=buffer;
        } else {
            fprintf(stderr, "Couldn't allocate buffer space\n");
            cstate->state=num_states;
            event_set( &cstate->ev, cstate->fd, EV_READ|EV_WRITE, conn_handle, state );
            event_add( &cstate->ev, NULL );

            return ;
        }
        
        cstate->datasize+=read( cstate->fd, cstate->buffer+cstate->datasize, cstate->buffsize-cstate->datasize );

        if( (needle=binbin(cstate->buffer, cstate->datasize, states[cstate->state].data,
			states[cstate->state].length ))!=NULL ) {
            cstate->datasize-=(needle-cstate->buffer)+states[cstate->state].length;
            memmove(cstate->buffer, needle+states[cstate->state].length, cstate->datasize );
            cstate->state++;
            event_set( &cstate->ev, cstate->fd, EV_READ|EV_WRITE, conn_handle, state );
        } else {
            /* Leave only state length bytes of the bytes read thus far in the buffer */
            memmove(cstate->buffer, cstate->buffer+cstate->datasize-states[cstate->state].length,
                    states[cstate->state].length );
            cstate->datasize=states[cstate->state].length;
        }

        event_add( &cstate->ev, NULL );
    }
    
}
#endif

void new_conn()
{
    struct conn_state *newconn=malloc(sizeof(struct conn_state));
    int error=FALSE;
    static unsigned short port=0;

    if( newconn==NULL ) {
	/* Couldn't even allocate the connection state */
	return;
    }

    newconn->buffer=NULL;
    newconn->buffsize=0;
    newconn->datasize=0;

    if( (newconn->fd=socket(PF_INET, SOCK_STREAM, 0))<0 ) {
	perror("socket creation failed");
	error=TRUE;
    }
    
    if( !error && fcntl( newconn->fd, F_SETFL, O_NONBLOCK )!=0 ) {
	perror("fcntl failed");
	error=TRUE;
    }

    if( !error && options.ports_start!=0 ) {
        struct sockaddr_in local;

        bzero( &local, sizeof( local ) );

        if( port<options.ports_start || port>=options.ports_end )
            port=options.ports_start;
        
        local.sin_family=AF_INET;
        local.sin_port=htons(port++);
        local.sin_addr=options.bind_addr;

        if( bind( newconn->fd, &local, sizeof(local) )!=0 ) {
            perror("local port binding failed");
            error=TRUE;
        }
    }

    if( !error && connect(newconn->fd, (struct sockaddr *)&options.address,
	    sizeof(options.address) )!=0 && errno!=EINPROGRESS) {
	fprintf(stderr, "%d\n", errno );
	perror("connect failed");
	error=TRUE;
    }

    newconn->state=0;

    if( !error ) {
#ifdef USE_EEL
        eel_group_id new_gid = eel_new_group_id();
        if (NULL == 
            eel_combine_OR(
              eel_add_read(new_gid, conn_handle_read, (void*)newconn, newconn->fd),
              eel_add_write(new_gid, conn_handle_write, (void*)newconn, newconn->fd)))
        {
            eel_delete_group_id(new_gid);
            perror("eel_combine_OR failed");
            error=TRUE;
        }
#else
	event_set( &newconn->ev, newconn->fd, EV_READ|EV_WRITE, conn_handle,
		newconn );

	if( event_add( &newconn->ev, NULL )!=0 ) {
	    perror("event_add failed");
	    error=TRUE;
	}
#endif
    }

    if( error ) {
	if( newconn->fd>=0 )
	    close( newconn->fd );
	free( newconn );
    }
}

int main(int argc, char *argv[])
{
    parse_cmdline(argc, argv);

#ifdef USE_EEL
    eel_group_id start_id = eel_new_group_id();
    init_libs(start_id);
    
    eel_dispatch_loop();
    
    eel_delete_group_id(start_id);
    eel_uninitialize();
#else
    init_libs();
    
    event_loop(0);
    
    /* free_libs(); */
#endif
    return 0;
}

