//
// Ryan Cunningham
//
// HOW TO USE:
// #include this file in breakpoint.c before the function
// definition for break_command. This file contains
// all the implementation functions necessary to support
// the ui commands listed below. You must also follow the
// instructions in gdb_initialize_breakpoint.c to add the
// functions to the ui command mappings. You can look at
// breakpoint.c in this directory (or use it).
//
// ALSO:
// There is one necessary change in infocmd.c. The function
// static void next_command (char *, int); must be made
// non-static so that we can call it from here
// Change the prototype at the top of the file
// and the definition later in the file. You can look at
// infocmd.c in this directory (or use it).
//
// FINALLY:
// Note that changing this file will not trigger a recompile
// of breakpoint.c because it is not listed as a dependency
// in the makefile. So, if you change this or gdb_initialize_breakpoint.c
// you will need to tap breakpoint.c in your gdb directory to
// get it to re-compile and link.
//

// headers
#include <regex.h>

//
//  modifications to support following callbacks
//
#define CBREAK_BUF_SIZE 256
#define CBREAK_MAX_FN_NAMES 5

// callback ui commands:
//  cbreak, cb
//  tcbreak, tcb
//  follow-callback, fol
//  tfollow-callback, tfol
//  set-follow-callback <pat> <arg num>
//  cnext, cn
// auto ui commands
//  acbreak, acb
//  atcbreak, atc
//  afollow-callback, afo
//  atfollow-callbak, atf
//  set-auto-follow-callbakc <pat>
//  anext, an

//
// warn - debug helper function
// leaving this set is not particularly verbose
// so do not worry about changing unless it gets
// too verbose for normal use
int g_bDebugFollowCallbacks = 1;
void warn(const char *format, ...)
{
  if (g_bDebugFollowCallbacks)
  { // output this string if debug flag is set
    va_list args;
    va_start(args, format);
    vprintf(format, args);
    va_end(args);
  }
}

//
// get_func_line - loads the symtab for a given function name
//
int
get_func_line (struct symtab_and_line *psal, char *fn_name)
{
  struct symtabs_and_lines sals;

  // check if symbols are loaded  
  if (!have_full_symbols () && !have_partial_symbols ())
    error ("No symbol table is loaded.  Use the \"file\" command.");

  // find the line information for the function name (symbol)
  sals = decode_line_1 (&fn_name, 0, 0, 0, 0, 0);
  if (!sals.nelts)
  { // ??? not sure when this case happens
    return -1;    /*  C++  */
  }
  if (sals.nelts > 1)
  {
    //ambiguous_line_spec (&sals); // this line would let the user choose 
                                   // between ambiguous symbol names, but is static
                                   // in another file
    xfree (sals.sals);
    return -1;
  }
  // take the first symbol from the line
  *psal = sals.sals[0];
  xfree (sals.sals);
  // check to make sure the location is valid
  if (psal->symtab == 0)
    error ("Could not find function %s.", fn_name);
  
  return 0;
}

//
// get_func_args - from a symtab and a line, load everything
// between the () preceeding -- which will be all of a functions
// arguments if the symtab corresponds to a function name lookup
// fill the buffer with this string
//
int
get_func_args (struct symtab *s, int line, char *buf, int buf_len)
{
  int c;
  int paren_pos = 0;
  int current_pos = 0;
  int func_pos = 0;
  int copy_args = 0;
  int desc;
  FILE *stream;

  // open up the source file
  desc = open_source_file (s);
  
  if (desc < 0)
    error ("Failed to open file.");
  
  if (s->line_charpos == 0)
    find_source_lines (s, desc);

  // check to make sure the line number is valid
  if (line < 1 || line > s->nlines)
  {
      close (desc);
      error ("Line number %d out of range; %s has %d lines.",
              line, s->filename, s->nlines);
  }

  // seek to the location of the line in the file
  if (lseek (desc, 0, 0) < 0)
  {
      close (desc);
      perror_with_name (s->filename);
  }
  
  // open a stream at this location
  stream = fdopen (desc, FOPEN_RB);
  clearerr (stream);
  
  // find '(' in the stream
  func_pos = s->line_charpos[line - 1];
  for (current_pos = 0; current_pos < func_pos; current_pos++) {
    c = fgetc (stream);
    if (c == EOF) 
    {
      fclose (stream);
      error("Reached end of file before finding char pos %d", func_pos);
    }
    if (c == '(')
    { // found it
      paren_pos = current_pos;
    }
  }
  
  // read to ')'
  if (fseek (stream, paren_pos + 1, SEEK_SET))
  {
    fclose (stream);
    error ("Failed to seek to paren pos = %d", paren_pos);
  }
  // copy the chars between '(' and ')'
  while (1)
  {
    c = fgetc (stream);
    if (c == EOF || c == ')' || copy_args >= (buf_len - 1)) {
      buf[copy_args] = 0;
      break; // done
    }
    if (c != '\n') // skip \n
    { // add the char to the buf
      buf[copy_args] = c;
      copy_args++;
    }
  }
  fclose (stream);
  return 0;
}

//
// find_arg - given a string args, and an argment number
// fill buf with the 0 based argument number from asrgs
//  use commas as argument separators (note: could be language dependent)
//
int
find_arg (char *buf, int buf_len, char *args, int arg_num)
{
  char *tok_comma, *tok_other, *tok_match = 0;
  char sep_comma[] = ",";
  char sep_other[] = " \t";
  
  buf[0] = 0;
  
  // make sure we have a string 
  if (!strlen (args))
    return -1;
    
  // find the arg_num comma or 0
  tok_comma = strtok (args, sep_comma);
  while (arg_num > 0 && tok_comma != NULL) 
  {
    arg_num--;
    tok_comma = strtok (0, sep_comma);
  }
  if (tok_comma) 
  { // we have the right comma, scan to end
    // to get the argument name
    tok_other = strtok (tok_comma, sep_other);
    while (tok_other)
    {
      tok_match = tok_other;
      tok_other = strtok (0, sep_other);
    }
    if (tok_match)
    { // matched the last token
      strncpy (buf, tok_match, buf_len);
      buf[buf_len-1] = 0;
      return 0;
    }
  }

  return -1;
}

//
// get_argument_name - given an function name, look up the source
// for that function definition and fill buf with the arg_num -th 
// argument name
//
int
get_argument_name (char *buf, int buf_len, char *fn_name, int arg_num)
{
  // need a symtab_and_line plus a buffer
  // to hold the full set of args
  struct symtab_and_line sal;
  char args[CBREAK_BUF_SIZE];
  
  args[0] = 0;
  buf[0] = 0;
  
  // get the line and symtab of the function
  if (get_func_line (&sal, fn_name))
  {
    warn("Failed to get_func_line\n");
    return -1;
  }
  // get the string of arguments of the function definiton from line
  if (get_func_args (sal.symtab, sal.line, args, CBREAK_BUF_SIZE))
  {
    warn("Failed to get_func_args\n");
    return -1;
  }
  // find the correct argument name from the argument string
  if (find_arg (buf, buf_len, args, arg_num))
  {
    warn("Failed to find_arg\n");
    return -1;
  }
  
  return 0;
}

//
// get_current_function_line - fill buf with the
// source line from the current line about to be
// executed (i.e. what is printed in the debugger
// when stepping through lines)
//
int
get_current_function_line (char *buf, int buf_len, int check_if_valid_line)
{
  struct frame_info *fi;
  struct symtab_and_line cursal;
  int c;
  int desc;
  int current_pos = 0;
  int paren_count = 0;
  FILE *stream;

  // use the stack frame to find the current line
  if (!target_has_stack)
  {
    if (!check_if_valid_line)
      error ("No stack.");
    return -1;
  }
  // find the current stack frame
  fi = get_current_frame ();

  if (fi == NULL)
  {
    if (!check_if_valid_line)
      error ("No stack after getting current frame.");
    return -1;
  }
  // get source and line info from the frame
  find_frame_sal (fi, &cursal);

  // open the source file for the current line
  desc = open_source_file (cursal.symtab);
  
  if (desc < 0)
  {
    if (!check_if_valid_line)
      error ("Failed to open current file.");
    return -1;
  }
  
  if (cursal.symtab->line_charpos == 0)
    find_source_lines (cursal.symtab, desc);
  // seek to the line
  if (lseek (desc, cursal.symtab->line_charpos[cursal.line - 1], 0) < 0)
  {
      close (desc);
      perror_with_name (cursal.symtab->filename);
  }
  // open a stream to read from the source
  stream = fdopen (desc, FOPEN_RB);
  clearerr (stream);
  // use parentheses to determine where the current
  // line ends - this assumes that the current line
  // has a function call. It will scan all the way to
  // the next function call if not, copying into buf
  for (current_pos = 0; current_pos < (buf_len - 1); ) 
  {
    c = fgetc (stream);
    if (c == EOF) 
    {
      fclose (stream);
      if (!check_if_valid_line)
        error ("Reached end of file before finishing current function line.");
      return -1;
    }
    if (c == '(')
    { // a new paren indicates nesting, keep track to find end
      paren_count++;
    }
    if (c == ')' && paren_count <= 1)
    { // we reached the closing paren, we are finished
      buf[current_pos] = ')';
      current_pos++;
      break;
    }
    else if (c == ')')
    { // this is the closing paren of a nested paren set
      paren_count--;
    }
    if (c != '\n')
    { // skip, newlines
      buf[current_pos] = c;
      current_pos++;
    }
  }
  fclose (stream);
  // make sure we null terminate, and have a string with some chars
  buf[current_pos] = 0;
  if (current_pos < (buf_len - 1))
    return 0;
  // unknown error
  return -1;
}

//
// fake_file_write - used to fake out a write
// later to collect the value being written
// these are used by get_expr_value
//
static char *temp_buf = NULL;
static long temp_buf_len = 0;
void
fake_file_write (struct ui_file *file,
                const char *buffer,
                long length_buffer)
{
  if (temp_buf && temp_buf_len)
  { // write to our buffer
    long min = min(temp_buf_len, length_buffer);
    strncpy (temp_buf, buffer, min);
    temp_buf[min] = 0;
  }
}

//
// get_expr_value - look up the value of the expression
// and return it in buf -- this will evaluate any
// expression the debugger can handle
//
int
get_expr_value (char *exp, char *buf, int buf_len)
{
  struct expression *expr;
  struct cleanup *old_chain = 0;
  struct value *val;
  int cleanup = 0;

  if (exp && *exp)
  { // prepare the expresion objects
    struct type *type;
    expr = parse_expression (exp);
    old_chain = make_cleanup (free_current_contents, &expr);
    cleanup = 1;
    // get the expression value
    val = evaluate_expression (expr);
  }
  else // we do not have a valid expression string
    return -1;

  // capture the value of the expression
  {
    struct ui_file *file = mem_fileopen (); // just use a mem file for now, there is no null_file ctor
    // set global bufs to our bufs
    temp_buf = buf;
    temp_buf_len = buf_len;
    // use the fake_file_write from above to capture
    // the output on the ui_file
    set_ui_file_write (file, fake_file_write);
    // pretty print the value
    value_print (val, file, 0, Val_pretty_default);
    // reset these to make sure they do not get accessed
    temp_buf = NULL;
    temp_buf_len = 0;
    // clean up the mem file
    ui_file_delete (file);
  }
  // clean up expr
  if (cleanup)
    do_cleanups (old_chain);
    
  return 0;
}

//
// fill_pattern - this is the big messy function
// that matches a source line against a pattern
// that can have * for wildcards. It really only
// matchs the first values for *, so the pattern
// is not a true regular expression.
// this really should be implemented using a reg exp
// package
// NOTE: this function is exceptionally long with
// poor readability...see notes in paper
int
fill_pattern (
     char *line,
     char *cb_fn_name, 
     int fn_len, 
     char *cb_arg_expr, 
     int arg_len,
     char *fn_pattern)
{
  char wildc = '*';
  const char *fn_pat = "<cb>";
  const char *id_pat = "<id>";
  const char *sep_pat = "<sep>";
  const char *seps = "),";
  const char *full_seps = ":()[]->.+-~!&*%/<>=^|?,";
  const char *space_like = " \n\t";
  char *fn_pat_loc = 0;
  char *id_pat_loc = 0;
  char *sep_pat_loc = 0;
  char *pat_pos = fn_pattern;
  char *wildcard_prev_pat_pos = 0;
  char *line_pos = 0;
  char *fn_copy = cb_fn_name;
  char *arg_copy = cb_arg_expr;
  int mode = 0;
  int got_fn = 0;
  int got_arg = 0;
  int just_id = (cb_arg_expr==NULL);
  
  line_pos = line;
  // find the cb and id tags
  fn_pat_loc = strstr (fn_pattern, fn_pat);
  id_pat_loc = strstr (fn_pattern, id_pat);
  sep_pat_loc = strstr (fn_pattern, sep_pat);
  if (fn_pat_loc == NULL || (id_pat_loc == NULL && !just_id))
  {
    printf_filtered ("Failed to find function or id markers in pattern.\n");
    return -1;
  }
  // just start scanning through the line
  for (; line_pos != NULL && pat_pos != NULL &&
         *line_pos != 0 && *pat_pos != 0;
       line_pos++)
  {
    //warn("mode=%d,line_pos=%s,pat_pos=%s\n", mode, line_pos, pat_pos);
  
    // NOTE: mode 0 = match exactly
    //       mode 1 = collect cb value
    //       mode 2 = collect id value
    //       mode -1 = in wildcard
    if (pat_pos == fn_pat_loc) { // pattern has reached the <cb> tag
      pat_pos += strlen (fn_pat);
      mode = 1;
      // skip spaces here to prevent <fn><sep> from stopping at a space
      while (*line_pos == ' ')
        line_pos++;
    }
    if (pat_pos == id_pat_loc) { // pattern has reached the <id> tag
      pat_pos += strlen (id_pat);
      mode = 2;
      // skip spaces here to prevent <id><sep> from stopping at a space
      while (*line_pos == ' ')
        line_pos++;
    }
    if (mode == 1) { // we are currently collecting the <cb> string
      if (*line_pos == *pat_pos || (pat_pos == sep_pat_loc && strchr(seps, *line_pos))) {
        *fn_copy = 0;
        got_fn = 1;
        if (*line_pos == *pat_pos) {
          pat_pos++; // this ended the arg collect and we had an exact match
        } else { // it was a sep that matched
          pat_pos += strlen (sep_pat); // move the pat forward
          sep_pat_loc = strstr (pat_pos, sep_pat); // find next <sep> if necessary
        }
        mode = 0;
        if (got_fn && (got_arg || just_id)) 
          return 0; // we have both cb and id, done (or just cb and do not need id)
      } else {
        if ((fn_copy - cb_fn_name) < fn_len) { // don't overrun the buffer
          if (*line_pos != ' ') { // trim while we are at it
            *fn_copy = *line_pos;
            fn_copy++;
          }
        } else {
          warn("Overran fn buffer.\n");
          return -1;
        }
      }
    } else if (mode == 2) { // we are currently collecting the <id> string
      if (*line_pos == *pat_pos || (pat_pos == sep_pat_loc && strchr(seps, *line_pos))) {
        *arg_copy = 0;
        got_arg = 1;
        if (*line_pos == *pat_pos) {
          pat_pos++; // this ended the arg collect and we had an exact match
        } else { // it was a sep that matched
          pat_pos += strlen (sep_pat); // move the pat forward
          sep_pat_loc = strstr (pat_pos, sep_pat); // find next <sep> if necessary
        }
        mode = 0;
        if (got_fn && got_arg) 
          return 0; // we have both cb and id, done
      } else {
        if ((arg_copy - cb_arg_expr) < arg_len) { // don't overrun the buffer
          if (*line_pos != ' ') { // trim while we are at it
            *arg_copy = *line_pos;
            arg_copy++;
          }
        } else {
          warn("Overran arg buffer.\n");
          return -1;
        }
      }
    } else if (mode == 0 && pat_pos == sep_pat_loc) {
      // pattern has reached a <sep> tag and we are in exact mode
      if (strchr(seps, *line_pos)) { // check if current char is a sep
        pat_pos += strlen (sep_pat); // move the pat forward
        sep_pat_loc = strstr (pat_pos, sep_pat); // find next <sep> if necessary
      } else {
        warn("<sep> did not exact match\n");
        return -1;
      }
    } else if (mode == -1 && pat_pos == sep_pat_loc && strchr(seps, *line_pos)) {
        pat_pos += strlen (sep_pat); // move the pat forward
        sep_pat_loc = strstr (pat_pos, sep_pat); // find next <sep> if necessary
        mode = 0; // now we are back to exact mode
    } else if ((mode == 0 || mode == -1) && *line_pos == *pat_pos) { // engage exact match
      if (mode == -1) {
        // we were just doing a wildcard, now hit a char that matches, need to remember this
        // spot in case we do not match
        wildcard_prev_pat_pos = pat_pos;
        mode = 0;
      }
      pat_pos++;
    } else if (*pat_pos == wildc) { // start a wildcard
      pat_pos++;
      mode = -1;
    } else if (mode == -1) { // in a wildcard
      // we are in wildcard, line_pos will just advance
    } else if (mode == 0 && strchr(space_like, *line_pos) && strchr(full_seps, *pat_pos)) {
      // we are in an exact match, right before a separator in the pattern (e.g. '(')
      // so, we need to allow spaces, or newlines, or tabs in the line match 
      // because these are legal
    } else {
      // got the wrong char, if we previously did a wildcard, set the pattern back
      // to where it was and try again.
      // NOTE: this is pretty weak because partial matches following a * will actually
      // throw off matching
      if (wildcard_prev_pat_pos) {
        pat_pos = wildcard_prev_pat_pos;
      } else {
        // got the wrong char and no wildcard
        return -1;
      }
    }
  }
  return -1;
}

//
// get_cb_name_and_arg_val - collect up the function
// name and argument expression value from the current line
// along with the current pattern
//
int
get_cb_names_and_arg_val (
     char cb_fn_names[CBREAK_MAX_FN_NAMES][CBREAK_BUF_SIZE],
     char *cb_arg_val, 
     int arg_len,
     regex_t *rg_pattern,
     int id_index,
     int check_if_valid_line)
{
  char current_line[CBREAK_BUF_SIZE];
  char arg_expr[CBREAK_BUF_SIZE];
  char *copy_to = NULL;
  current_line[0] = 0;
  arg_expr[0] = 0;
  int number_of_cb_names = 0;
  size_t number_of_matches = min(rg_pattern->re_nsub+1,CBREAK_MAX_FN_NAMES);
  regmatch_t *match_result = NULL;
  int match_index = 0;
  
  if ((match_result = (regmatch_t *) malloc(sizeof(regmatch_t) * number_of_matches)) == NULL)
  {
    error ("Out of memory, malloc failed.");
  }
  
  // get the current line
  if (get_current_function_line (current_line, sizeof(current_line), check_if_valid_line))
  {
    free((void*)match_result);
    if (!check_if_valid_line)
      error ("Failed to get current function line.");
    return 0;
  }
  
  //warn("Current line = %s, number of matches = %d, id index = %d\n", current_line, number_of_matches, id_index);
  // find the function names and arg name
  // use the regex pattern to find the function and arg name
  if (regexec(rg_pattern, current_line, number_of_matches, match_result, 0) != 0) 
  {
    free((void*)match_result);
    if (!check_if_valid_line)
      error("Pattern did not match!\n");
    return 0;
  }
  // start with index 1 to skip the main non-grouped regex match
  for (match_index = 1; match_index < number_of_matches; match_index++)
  {
    //warn("Index = %d, max = %d\n", match_index, number_of_matches);
    copy_to = arg_expr;
    if (match_index != id_index)
    {
      copy_to = cb_fn_names[number_of_cb_names];
      number_of_cb_names++;
    }
    if (match_result[match_index].rm_eo > match_result[match_index].rm_so && 
        *(current_line+match_result[match_index].rm_so) == ',')
    { // this is a bogus entry, use the empty string
      copy_to[0] = 0;
    }
    else
    {
      strncpy(copy_to, 
              current_line+match_result[match_index].rm_so, 
              match_result[match_index].rm_eo-match_result[match_index].rm_so);
      copy_to[match_result[match_index].rm_eo-match_result[match_index].rm_so] = 0;
    }
  }
  free((void*)match_result);
  // evaluate the arg as an expression
  if (get_expr_value (arg_expr, cb_arg_val, arg_len))
  {
    if (!check_if_valid_line)
      error ("Failed to get expression value for\n%s\nin\n%s.", arg_expr, current_line);
    return 0;
  }
    
  return number_of_cb_names;
}

//
// store the pattern and argument number
// between calls
static int perm_arg_num = 0;
static int init_pattern = 1;
static char perm_fn_pattern[CBREAK_BUF_SIZE];
static regex_t perm_regex_pattern;
static int perm_id_index = 1;

//
// compile_regex
// take a fn_pattern (with <id> and [<cb>]+ tags), convert and compile
// to a regex_t
//
int
compile_regex (char *fn_pattern, regex_t *rg_pattern, int *id_index)
{
  static const char *C_identifier_pattern = " *\\([A-z][A-z,0-9,_]*\\)";
  const int C_identifier_len = strlen(C_identifier_pattern);
  static const char *exp_pattern = " *\\([^,]\\+\\)"; // everything but comma
  const int exp_pattern_len = strlen(exp_pattern);
  int regcomp_res = 0;
  int new_id_index = 1;
  int found_id = 0;
  char real_pattern[CBREAK_BUF_SIZE*4];
  char *cur_pat_str = fn_pattern;
  char *cur_real_str = real_pattern;
  
  // find the first tags
  char *id_str = strstr(cur_pat_str, "<id>");
  char *cb_str = strstr(cur_pat_str, "<cb>");
  if (id_str == NULL || cb_str == NULL || rg_pattern == NULL || id_index == NULL) 
  {
    return -1;
  }
  // replace tags with C 'identifier' regex pattern
  // and find the index of the <id> tag
  while (cb_str)
  {
    //warn("cur_pat_str = %s, cb_str = %s\n", cur_pat_str, cb_str);
    if (found_id == 0 && id_str < cb_str) 
    {
      // copy everything up to the tag into the real pattern
      strncpy(cur_real_str, cur_pat_str, id_str-cur_pat_str);
      cur_real_str += id_str-cur_pat_str;
      cur_pat_str = id_str;
      // substitute an exp pattern for the <id> tag
      strncpy(cur_real_str, exp_pattern, exp_pattern_len);
      cur_real_str += exp_pattern_len;
      // we have found the <id> tag
      found_id = 1;
      // move past the <id> tag
      cur_pat_str += 4;
    }
    else
    {
      if (found_id == 0) {
        // we have not found the <id> tag, but another <cb> has gone by
        // move the id index up one
        new_id_index++;
      }
      // copy everything up to the tag into the real pattern
      strncpy(cur_real_str, cur_pat_str, cb_str-cur_pat_str);
      cur_real_str += cb_str-cur_pat_str;
      cur_pat_str += cb_str-cur_pat_str;
      // substitute a C identifier pattern for callback function names
      strncpy(cur_real_str, C_identifier_pattern, C_identifier_len);
      cur_real_str += C_identifier_len;
      // move past the current <cb> tag and find the next
      cur_pat_str += 4;
      cb_str = strstr(cur_pat_str, "<cb>");
    }
  }
  // copy everything up to the tag into the real pattern
  strncpy(cur_real_str, cur_pat_str, strlen(cur_pat_str));
  cur_real_str += strlen(cur_pat_str);
  // NULL terminate the real pattern string
  *cur_real_str = '\0';
  
  //warn("Real pattern = %s\n", real_pattern);
  
  // compile the actual pattern
  regcomp_res = regcomp(rg_pattern, real_pattern, 0);
  if (regcomp_res == 0) {
    *id_index = new_id_index;
    return 0;
  }
  
  // debugging of default pattern
  //{
  //  char reg_error_message[CBREAK_BUF_SIZE];
  //  regerror(regcomp_res, rg_pattern, reg_error_message, CBREAK_BUF_SIZE);
  //  warn ("regcomp failed to compile regex: %s\nreal pattern = %s\n", reg_error_message, real_pattern);
  //}
  
  return regcomp_res;
}

//
// check_perm_pattern - make sure the permanent
// pattern is filled with default if it has not 
// been initialized
//
void
check_perm_pattern ()
{
  if (init_pattern)
  { // default pattern
    init_pattern = 0;
    perm_arg_num = 0;
    strcpy (perm_fn_pattern, "eel_add_[A-z,0-9,_]* *(<id>,<cb> *\\(,<cb> *\\)\\{0,1\\}[,)]");
    compile_regex (perm_fn_pattern, &perm_regex_pattern, &perm_id_index);
  }
}

//
// reset_perm_pattern - reset the permanent pattern
// back to the default
//
void
reset_perm_pattern ()
{
  if (init_pattern == 0) {
    init_pattern = 1;
    regfree(&perm_regex_pattern);
  }
  check_perm_pattern ();
}

//
// store the auto pattern between calls
static int init_auto_pattern = 1;
static char perm_auto_fn_pattern[CBREAK_BUF_SIZE];

//
// check_auto_perm_pattern - make sure the auto permanent
// pattern is filled with default if it has not 
// been initialized
//
void
check_auto_perm_pattern ()
{
  if (init_auto_pattern)
  { // default pattern
    init_auto_pattern = 0;
    strcpy (perm_auto_fn_pattern, "*wrap(<cb><sep>");
  }
}

//
// reset_auto_perm_pattern - reset the permanent pattern
// back to the default
//
void
reset_auto_perm_pattern ()
{
  init_auto_pattern = 1;
  check_auto_perm_pattern ();
}

//
// store the check_expr between calls
static int init_auto_check_expr = 1;
static char perm_next_id_expr[CBREAK_BUF_SIZE];
static char perm_current_id_expr[CBREAK_BUF_SIZE];

//
// check_auto_perm_check_id_exprs - make sure the permanent
// next and current id exprs are filled with defaults if they
// have not been initialized
//
void
check_auto_perm_check_id_exprs ()
{
  if (init_auto_check_expr)
  { // default pattern
    init_auto_check_expr = 0;
    strcpy (perm_next_id_expr, "debug_wrap_id::next_id");
    strcpy (perm_current_id_expr, "debug_wrap_id::flag_break && debug_wrap_id::current_id");
  }
}

//
// reset_auto_perm_check_id_exprs - reset the permanent 
// next and current id exprs back to the defaults
//
void
reset_auto_perm_check_id_exprs ()
{
  init_auto_check_expr = 1;
  check_auto_perm_check_id_exprs ();
}

//
// parse_callback_arg - parse a string of the form
// <callback function pattern> <argument number> and fill
// pattern and parg_num with the values
//
void
parse_callback_arg (char *arg, char *pattern, int pat_len, int *parg_num, regex_t *rg_pattern, int *id_index)
{
  int remove_len = 0;
  int new_arg_num = 0;
  char *new_pattern = NULL;
  regex_t new_rg_pattern;
  int new_id_index = 0;
  int reg_err_no = 0;
  
  if (arg == NULL || strlen(arg) <= 0)
  { // hack to reset pattern
    reset_perm_pattern ();
    printf_filtered ("Pattern has been reset to default = %d %s\n", perm_arg_num, perm_fn_pattern);
    return;
  }
  if (arg && strlen (arg) >= pat_len)
    error ("Argument is too long.");
    
  if (sscanf (arg, "%d %n", &new_arg_num, &remove_len) < 1)
  {
    error ("Bad pattern. Pattern should be of the form '<argument number> <callback function pattern>'.");
  }
  new_pattern = arg+remove_len;
  
  if (strstr (new_pattern, "<cb>") == NULL || 
      strstr (new_pattern, "<id>") == NULL)
  {
    error ("Pattern does not contain '<cb>' and '<id>'.");
  }
  
  // compile the regex pattern
  reg_err_no = compile_regex(new_pattern, &new_rg_pattern, &new_id_index);
  if (reg_err_no != 0) 
  {
    char reg_error_message[CBREAK_BUF_SIZE];
    regerror(reg_err_no, &new_rg_pattern, reg_error_message, CBREAK_BUF_SIZE);
    regfree(&new_rg_pattern);
    error ("Failed to compile regex: %s", reg_error_message);
  }
  regfree(&new_rg_pattern);
  // save the new pattern and arg number
  *parg_num = new_arg_num;
  strcpy(pattern, new_pattern);
  // compile the actual pattern
  regfree(rg_pattern);
  compile_regex(pattern, rg_pattern, id_index);
  // tell the user we are all set
  printf_filtered ("Pattern = '%s', argument number = %d\n", pattern, *parg_num);
}

//
// set_callback_arg - use the arg to set the permanent pattern
//
void
set_callback_arg (char *arg)
{
  check_perm_pattern ();
  parse_callback_arg (arg, perm_fn_pattern, sizeof(perm_fn_pattern), &perm_arg_num, &perm_regex_pattern, &perm_id_index);
}

//
// set_acallback_arg - use the arg to set the auto permanent pattern
//
void
set_acallback_arg (char *arg)
{
  check_auto_perm_pattern ();
  if (arg && strlen(arg)) {
    strncpy (perm_auto_fn_pattern, arg, sizeof(perm_auto_fn_pattern));
  } else {
    reset_auto_perm_pattern ();
    printf_filtered ("Auto pattern has been reset to default = %s\n", perm_auto_fn_pattern);
  }
}

//
// set_id_exprs - use the arg to set the permanent id exprs
//
void
set_id_exprs (char *arg)
{
  check_auto_perm_check_id_exprs ();
  if (!arg || !strlen(arg) || 
      (sscanf(arg, "%s %s", perm_next_id_expr, perm_current_id_expr) != 2)) 
  {
    reset_auto_perm_check_id_exprs ();
    printf_filtered ("Auto id expressions are set to default = %s and %s\n", perm_next_id_expr, perm_current_id_expr);
  }
}

//
// cbreak_command_flag - the main command handler
// this can create a cbreak or a tcbreak,
// handles the arguments as a possible temporary 
// pattern. It can also check if a line matches.
//
int
cbreak_command_flag (char *arg, int from_tty, int flag, int check_if_valid_line)
{
  char *fn_pattern = perm_fn_pattern;
  regex_t *rg_pattern = &perm_regex_pattern;
  regex_t new_rg_pattern;
  int regfree_new_rg_pattern = 0;
  int new_id_index = 0;
  int id_index = perm_id_index;
  int arg_num = perm_arg_num;
  int number_of_cb_names = 0;
  int i = 0;
  // these buffers are used primarily 
  // to make the debugging easier
  // between steps
  char cb_fn_names[CBREAK_MAX_FN_NAMES][CBREAK_BUF_SIZE];
  char arg_value[CBREAK_BUF_SIZE];
  char arg_name[CBREAK_BUF_SIZE];
  char command[CBREAK_BUF_SIZE];
  char pattern[CBREAK_BUF_SIZE];

  check_perm_pattern ();
  
  arg_value[0]  = 0;
  arg_name[0]   = 0;
  command[0]    = 0;
  pattern[0]    = 0;
  
  // parse the arg if necessary
  if (arg && strlen (arg))
  {
    regcomp(&new_rg_pattern, "", 0); // compile a pattern so new_rg_pattern is valid
    parse_callback_arg (arg, pattern, sizeof(pattern), &arg_num, &new_rg_pattern, &new_id_index);
    fn_pattern = pattern;
    rg_pattern = &new_rg_pattern;
    id_index = new_id_index;
    regfree_new_rg_pattern = 1;
  }
  
  // get the cb function name and argument value
  number_of_cb_names = 
    get_cb_names_and_arg_val (cb_fn_names, 
                              arg_value, 
                              sizeof(arg_value), 
                              rg_pattern,
                              id_index,
                              check_if_valid_line);
  if (regfree_new_rg_pattern) {
    regfree(&new_rg_pattern);
  }

  // if we are just checking to see if this line matches, we can answer that now
  if (check_if_valid_line)
  {
    if (number_of_cb_names > 0)
      return 0;
    else
      return -1;
  }
  
  // loop on callback function names to actually create the breakpoints
  for (i = 0; i < number_of_cb_names; i++)
  {
    if (strlen(cb_fn_names[i]))
    {
      // get the argument name
      if (get_argument_name (arg_name, sizeof(arg_name), cb_fn_names[i], arg_num))
        error ("Failed to get argument name from function name %s, argument number %d", cb_fn_names[i], arg_num);
      // build the command
      sprintf (command, "%s if (%s==%s)", cb_fn_names[i], arg_name, arg_value);
      warn("%s\n", command);
      // set the breakpoint
      break_command_1 (command, flag, from_tty, NULL);
    }
  }
  return 0;
}

//
// acbreak_command_flag - the main command handler
// this can create an auto acbreak and atcbreak,
// handles the arguments as a possible temporary 
// pattern. 
//
int
acbreak_command_flag (char *arg, int from_tty, int flag, int check_if_valid_line)
{ // set an auto callback break
  char *fn_pattern = perm_auto_fn_pattern;
  char current_line[CBREAK_BUF_SIZE];
  char command[CBREAK_BUF_SIZE];
  char cb_fn_name[CBREAK_BUF_SIZE];
  char expr_value[CBREAK_BUF_SIZE];
  
  check_auto_perm_pattern ();
  check_auto_perm_check_id_exprs ();
  
  current_line[0] = 0;
  command[0]      = 0;
  cb_fn_name[0]   = 0;
  expr_value[0]   = 0;
  
  // use the arg if necessary
  if (arg && strlen (arg))
  {
    fn_pattern = arg;
  }
  
  // get the current line
  if (get_current_function_line (current_line, sizeof(current_line), check_if_valid_line))
  {
    if (!check_if_valid_line)
      error ("Failed to get current function line.");
    return -1;
  }
  
  // get the cb function name
  if (fill_pattern (current_line, cb_fn_name, sizeof(cb_fn_name), NULL, 0, fn_pattern))
  {
    if (!check_if_valid_line)
      error ("Failed to fill pattern.\nLine = %s\nPattern = '%s'", current_line, fn_pattern);
    return -1;
  }
  
  if (!check_if_valid_line)
  {
    // find out what the next debug id will be
    if (get_expr_value (perm_next_id_expr, expr_value, sizeof(expr_value)))
      error("Failed to get expression value for debug_wrap_next_id, did you include debug_wrap.h?\n");

    sprintf(command, "%s if (%s==%s)", cb_fn_name, perm_current_id_expr, expr_value);
    warn("%s\n", command);
    break_command_1 (command, flag, from_tty, NULL);
  }
  return 0;
}

void
cbreak_command (char *arg, int from_tty)
{ // just set the flag to permanant
  cbreak_command_flag (arg, from_tty, 0, 0);
}

void
tcbreak_command (char *arg, int from_tty)
{ // just set the flag to temporary
  cbreak_command_flag (arg, from_tty, BP_TEMPFLAG, 0);
}

void
follow_callback (char *arg, int from_tty)
{ // just set a cbreak then continue
  cbreak_command_flag (arg, from_tty, 0, 0);
  continue_command (arg, from_tty); // *** do we want to clear these out?
}

void
tfollow_callback (char *arg, int from_tty)
{ // just set a tcbreak then continue
  cbreak_command_flag (arg, from_tty, BP_TEMPFLAG, 0);
  continue_command (arg, from_tty); // *** do we want to clear these out?
}

void
set_follow_callback (char *arg, int from_tty)
{ // set the permanent pattern
  set_callback_arg (arg);
}

// this function was exported infocmd.c for external access
// NOTE: infocmd.c must be modified to 
extern void next_command (char *, int);
// use it to execute the "next" command
void
callback_next (char *arg, int from_tty)
{ 
  // see if we are on a auto-next callback line
  if (cbreak_command_flag (arg, from_tty, BP_TEMPFLAG, 1))
  { // just do a next
    next_command (arg, from_tty);
  }
  else
  { // set the breakpoint and continue
    printf_filtered("Following callback\n");
    cbreak_command_flag (arg, from_tty, BP_TEMPFLAG, 0);
    continue_command (arg, from_tty); // *** do we want to clear these out?
  }
}

void
set_afollow_callback (char *arg, int from_tty)
{ // set the auto permanent pattern
  set_acallback_arg (arg);
}

void
set_auto_id_exprs (char *arg, int from_tty)
{ // set the auto permanent id exprs
  set_id_exprs (arg);
}

void
acbreak_command (char *arg, int from_tty)
{ // just set the flag to permanant
  acbreak_command_flag (arg, from_tty, 0, 0);
}

void
atcbreak_command (char *arg, int from_tty)
{ // just set the flag to temporary
  acbreak_command_flag (arg, from_tty, BP_TEMPFLAG, 0);
}

void 
afollow_callback (char *arg, int from_tty)
{ // just set a acbreak then continue
  acbreak_command_flag (arg, from_tty, 0, 0);
  continue_command (arg, from_tty); // *** do we want to clear these out?
}

void 
atfollow_callback (char *arg, int from_tty)
{ // just set a atcbreak then continue
  acbreak_command_flag (arg, from_tty, BP_TEMPFLAG, 0);
  continue_command (arg, from_tty); // *** do we want to clear these out?
}

void
auto_next (char *arg, int from_tty)
{ 
  // see if we are on a auto-next callback line
  if (acbreak_command_flag (arg, from_tty, BP_TEMPFLAG, 1))
  { // just do a next
    next_command (arg, from_tty);
  }
  else
  { // set the breakpoint and continue
    printf_filtered("Auto following callback\n");
    acbreak_command_flag (arg, from_tty, BP_TEMPFLAG, 0);
    continue_command (arg, from_tty); // *** do we want to clear these out?
  }
}
//
// End cbreak changes (more in _initialize_breakpoint())
//
