//
// C Implementation: eel_epoll
//
// Description: 
//
//
// Author: Ryan Cunningham <rcunning@cs.ucla.edu>, (C) 2004
//
// Copyright: Ryan Cunningham
//
//

// a mess of headers here, likely can remove some
#include <sys/types.h>
#include <tree.h>
#include <queue.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <err.h>
#include <assert.h>

#include <unistd.h>
#include <sys/resource.h>
#include <sys/epoll.h>

#include "eel.h"
#include "eel_util.h"
#include "eel_epoll.h"

//--- check EEL_NO_EPOLL macro

// prototypes for eel_operators implementation
int   epoll_init       ();
void  epoll_end        ();
int   epoll_add        (struct eel_lib_struct *eel, struct eel_event *e);
int   epoll_remove     (struct eel_lib_struct *eel, struct eel_event *e);
int   eel_epoll_wait   (struct eel_lib_struct *eel, int timeout_milliseconds);

// eel operator instance
struct eel_operators epoll_operators = {
  "epoll",
  &epoll_init,
  &epoll_end,
  &epoll_add,
  &epoll_remove,
  &eel_epoll_wait
};

// --------------------------------------------------------
// get_epoll_eel_operators - return our operators
// --------------------------------------------------------
struct eel_operators *get_epoll_eel_operators()
{
  if (getenv("EEL_NO_EPOLL")) {
    return NULL;
  }
  return &epoll_operators;
}

// helper data structure -- need to be able to
// index by fd so red/write/error event can be merged
struct eel_events {
  struct eel_event *read;
  struct eel_event *write;
  struct eel_event *error;
};

// structure used to save data between eel operator calls
struct epoll_data_struct {
  struct epoll_event *epoll_array;
  int                epoll_count;
  struct eel_events  *events_array;
  int                events_count;
  int                epoll_fd;
} epoll_data;

//
// Implementation
//

// --------------------------------------------------------
// epoll_init - initialize
// --------------------------------------------------------
int epoll_init()
{
  int file_limit = 32000; // default
  struct rlimit rl;
  // clear out our structure
  memset(&epoll_data, 0, sizeof(struct epoll_data_struct));
  // get the max files
  if (getrlimit(RLIMIT_NOFILE, &rl) == 0 && rl.rlim_cur != RLIM_INFINITY) {
    file_limit = rl.rlim_cur;
  }
  // inform epoll that we want to handle about file_limit many fds
  if ((epoll_data.epoll_fd = epoll_create(file_limit)) == -1) {
    EEL_DEBUG_ERRORNO_MESSAGE("epoll_create failed")
    return -1;
  }
  // create our arrays
  epoll_data.epoll_array = (struct epoll_event*)
    malloc(file_limit*sizeof(struct epoll_event));
  if (epoll_data.epoll_array == NULL) {
    EEL_DEBUG_MESSAGE1("malloc failed when creating epoll_array of size %d", file_limit)
    return -1;
  }
  epoll_data.epoll_count = file_limit;
  
  epoll_data.events_array = (struct eel_events*)
    calloc(file_limit, sizeof(struct eel_events));
  if (epoll_data.events_array == NULL) {
    EEL_DEBUG_MESSAGE1("calloc failed when creating events_array of size %d", file_limit)
    return -1;
  }
  epoll_data.events_count = file_limit;
  
  return 0;
}

// --------------------------------------------------------
// epoll_end - clean up
// --------------------------------------------------------
void epoll_end()
{
  // free the memory created for the arrays
  SAFE_FREE(epoll_data.epoll_array);
  SAFE_FREE(epoll_data.events_array);
  // close the epoll fd
  if (epoll_data.epoll_fd) {
    close(epoll_data.epoll_fd);
  }
}

// --------------------------------------------------------
// update_events_size - realloc if we need more room in
// the events_array, return -1 on error, 0 on success
// -------------------------------------------------------
static int update_events_size(int new_events_count_requested)
{
  struct eel_events *new_events_array = NULL;
  int new_events_count = epoll_data.events_count;
  
  if (new_events_count_requested > epoll_data.events_count) {
    // double size until we are above the requested size
    while (new_events_count < new_events_count_requested) {
      new_events_count <<= 1;
    }
    // realloc
    new_events_array = (struct eel_events*)
      realloc(epoll_data.events_array, new_events_count*sizeof(struct eel_events));
    if (new_events_array == NULL) {
      epoll_data.events_array = NULL;
      epoll_data.events_count = 0;
      EEL_DEBUG_MESSAGE("Failed to realloc in epoll update_events_size.")
      return -1;
    }
    // make sure the new memory is zeroed
    if (new_events_count > epoll_data.events_count) {
      // clear out the new memory
      memset(new_events_array + (epoll_data.events_count * sizeof(struct eel_events)), 0, 
        (new_events_count - epoll_data.events_count) * sizeof(struct eel_events));
    }
    // set our structure to point to the new data
    epoll_data.events_array = new_events_array;
    epoll_data.events_count = new_events_count;
  }
  
  return 0;
}

// --------------------------------------------------------
// get_epoll_events - helper function that returns the 
// the correstponding epoll event from an eel_event
// --------------------------------------------------------
static int get_epoll_event(struct eel_event *e)
{
  int return_events = 0;
  if (e) {
    switch (e->type) {
    case EEL_READ_EVENT:
      return_events = EPOLLIN;
      break;
    case EEL_WRITE_EVENT:
      return_events = EPOLLOUT;
      break;
    case EEL_ERROR_EVENT:
      return_events = (EPOLLERR|EPOLLHUP);
      break;
    case EEL_TIMER_EVENT:
      assert(1);
      break;
    };
  }
  return return_events;
}

// --------------------------------------------------------
// reset_eel_events_by_type - helper function to deal with
// setting the corresponding eel_events field by type
// --------------------------------------------------------
static void reset_eel_events_by_type(struct eel_events *pevents, int type, struct eel_event *set_to)
{
  switch (type) {
  case EEL_READ_EVENT:
    pevents->read = set_to;
    break;
  case EEL_WRITE_EVENT:
    pevents->write = set_to;
    break;
  case EEL_ERROR_EVENT:
    pevents->error = set_to;
    break;
  case EEL_TIMER_EVENT:
    assert(1);
    break;
  };
}

// --------------------------------------------------------
// epoll_add - an event has been added
// --------------------------------------------------------
int epoll_add(struct eel_lib_struct *eel, struct eel_event *e)
{
  struct eel_events *pevents = NULL;
  struct epoll_event epoll_event;
  int epoll_operator;
  
  // make sure we have enough room
  if (e->fd >= epoll_data.events_count) {
    // need more space for events
    if (update_events_size(e->fd) == -1) {
      return -1;
    }
  }
  // update or add the event, merge the event flags
  pevents = &epoll_data.events_array[e->fd];
  epoll_event.events = 
    get_epoll_event(e) | 
    get_epoll_event(pevents->read) | 
    get_epoll_event(pevents->write) | 
    get_epoll_event(pevents->error);
  epoll_operator = ( (pevents->read || pevents->write || pevents->error) ? EPOLL_CTL_MOD : EPOLL_CTL_ADD);
  epoll_event.data.ptr = (void*)pevents;
  
  // perform the epoll operation
  if (epoll_ctl(epoll_data.epoll_fd, epoll_operator, e->fd, &epoll_event) == -1) {
    EEL_DEBUG_ERRORNO_MESSAGE("epoll_ctl failed in add")
    return -1;
  }
  
  // update our events_array
  reset_eel_events_by_type(pevents, e->type, e);
  
  return 0;
}

// --------------------------------------------------------
// epoll_remove - an event has been removed
// --------------------------------------------------------
int epoll_remove(struct eel_lib_struct *eel, struct eel_event *e)
{
  struct eel_events *pevents = NULL;
  struct epoll_event epoll_event;
  int epoll_operator;
  
  if (e->fd > epoll_data.events_count) {
    // out of range, ignore
    return 0;
  }
  // figure out the operator and remaining events
  pevents = &epoll_data.events_array[e->fd];
  epoll_event.events = 
    (~get_epoll_event(e)) & 
    (get_epoll_event(pevents->read) | 
     get_epoll_event(pevents->write) | 
     get_epoll_event(pevents->error));
  epoll_operator = ( epoll_event.events == 0 ? EPOLL_CTL_DEL : EPOLL_CTL_MOD);
  epoll_event.data.ptr = (void*)pevents;
  
  // have to clear out the event no matter
  // what, because the eel_event will be
  // gone after this call
  reset_eel_events_by_type(pevents, e->type, NULL);
  
  // perform the epoll operation
  if (epoll_ctl(epoll_data.epoll_fd, epoll_operator, e->fd, &epoll_event) == -1) {
    EEL_DEBUG_ERRORNO_MESSAGE("epoll_ctl failed in remove")
    return -1;
  }
  
  return 0;
}

// --------------------------------------------------------
// eel_epoll_wait - implement wait functionality, renamed
// to avoid conflict with epoll sys calls
// --------------------------------------------------------
int eel_epoll_wait(struct eel_lib_struct *eel, int timeout_milliseconds)
{
  int epoll_result = 0;
  int i, epoll_events;
  struct eel_events *pevents = NULL;

  // just go ahead and do the wait -- everything should be set up
  epoll_result = epoll_wait(epoll_data.epoll_fd, 
                            epoll_data.epoll_array, 
                            epoll_data.epoll_count,
                            timeout_milliseconds);
  // add the results to the fire queue
  for (i = 0; i < epoll_result; i++) {
    epoll_events = epoll_data.epoll_array[i].events;
    pevents = (struct eel_events*)epoll_data.epoll_array[i].data.ptr;
    // check and see which of these events fired, then add them
    if ( ((epoll_events & EPOLLIN) || (epoll_events & EPOLLPRI)) &&
         pevents->read)
    {
      eel->add_to_fire(pevents->read);
    }
    if ( (epoll_events & EPOLLOUT) &&
         pevents->write)
    {
      eel->add_to_fire(pevents->write);
    }
    if ( ((epoll_events & EPOLLERR) || (epoll_events & EPOLLHUP)) &&
         pevents->error)
    {
      eel->add_to_fire(pevents->error);
    }
  }
  
  return epoll_result;
}
