//
// C Implementation: eel_poll
//
// Description: 
//
//
// Author: Ryan Cunningham <rcunning@cs.ucla.edu>, (C) 2004
//
// Copyright: Ryan Cunningham
//
//

// a mess of headers here, likely can remove some
#include <sys/types.h>
#include <tree.h>
#include <queue.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <err.h>
#include <assert.h>

#include <sys/poll.h>

#include "eel.h"
#include "eel_util.h"
#include "eel_poll.h"

//--- check EEL_NO_POLL macro

// prototypes for eel_operators implementation
int   poll_init       ();
void  poll_end        ();
int   poll_add        (struct eel_lib_struct *eel, struct eel_event *e);
int   poll_remove     (struct eel_lib_struct *eel, struct eel_event *e);
int   poll_wait       (struct eel_lib_struct *eel, int timeout_milliseconds);

// eel operator instance
struct eel_operators poll_operators = {
  "poll",
  &poll_init,
  &poll_end,
  &poll_add,
  &poll_remove,
  &poll_wait
};

// --------------------------------------------------------
// get_poll_eel_operators - return our operators
// --------------------------------------------------------
struct eel_operators *get_poll_eel_operators()
{
  if (getenv("EEL_NO_POLL")) {
    return NULL;
  }
  return &poll_operators;
}

// structure used to save data between eel operator calls
struct poll_data_struct {
  int              current_event_count;  // running count of events
  struct pollfd    *poll_array;          // array of pollfd structs for poll()
  struct eel_event **event_array;        // numerically matching array of pointers to eel_event
  int              poll_count;           // the size of the two arrays above
} poll_data;

//
// Implementation
//

// --------------------------------------------------------
// poll_init - initialize
// --------------------------------------------------------
int poll_init()
{
  // clear out our structure
  memset(&poll_data, 0, sizeof(struct poll_data_struct));
  return 0;
}

// --------------------------------------------------------
// poll_end - clean up
// --------------------------------------------------------
void poll_end()
{
  // free the memory created for the arrays
  SAFE_FREE(poll_data.poll_array);
  SAFE_FREE(poll_data.event_array);
}

// --------------------------------------------------------
// poll_add - an event has been added
// --------------------------------------------------------
int poll_add(struct eel_lib_struct *eel, struct eel_event *e)
{
  // just track the number of events
  poll_data.current_event_count++;
  return 0;
}

// --------------------------------------------------------
// poll_remove - an event has been removed
// --------------------------------------------------------
int poll_remove(struct eel_lib_struct *eel, struct eel_event *e)
{
  // just track the number of events
  poll_data.current_event_count--;
  return 0;
}

// --------------------------------------------------------
// poll_update_size - helper function to grow the arrays
// --------------------------------------------------------
static int poll_update_size()
{
  // if we have more events than space, grow the arrays
  if (poll_data.current_event_count > poll_data.poll_count)
  {
    // grow the array by *2 when necessary to prevent
    // growing too many times as event counts increase
    if (poll_data.poll_count < 32) {
      poll_data.poll_count = 32;
    } else {
      poll_data.poll_count *= 2;
    }
    // realloc to the new size
    poll_data.poll_array  = (struct pollfd*)
      realloc(poll_data.poll_array, poll_data.poll_count*sizeof(struct pollfd));
    poll_data.event_array = (struct eel_event**)
      realloc(poll_data.event_array, poll_data.poll_count*sizeof(struct eel_event*));  
    if (poll_data.poll_array == NULL ||
        poll_data.event_array == NULL) {
        EEL_DEBUG_MESSAGE("realloc() returned NULL in poll_update_size.")
        return -1;
    }
  }
  return 0;
}

// --------------------------------------------------------
// get_poll_events - helper function that returns the 
// the correstponding poll event from an eel_event
// --------------------------------------------------------
static int get_poll_event(struct eel_event *e)
{
  int return_events = 0;
  if (e) {
    switch (e->type) {
    case EEL_READ_EVENT:
      return_events = POLLIN;
      break;
    case EEL_WRITE_EVENT:
      return_events = POLLOUT;
      break;
    case EEL_ERROR_EVENT:
      return_events = (POLLERR|POLLHUP);
      break;
    case EEL_TIMER_EVENT:
      assert(1);
      break;
    };
  }
  return return_events;
}

// --------------------------------------------------------
// poll_wait - implement wait functionality
// --------------------------------------------------------
int poll_wait(struct eel_lib_struct *eel, int timeout_milliseconds)
{
  int poll_result = 0;
  int fd_count = 0;
  struct eel_event *e;
  
  // update the array sizes
  if (poll_update_size() == -1) {
    return -1;
  }
  // fill the arrays with the correct fds and events, note: O(n)
  TAILQ_FOREACH(e, &eel->queue, queue_entry) {
    assert(fd_count < poll_data.poll_count);
    // fill in the common parts of the struct
    struct pollfd *pfd = &poll_data.poll_array[fd_count];
    pfd->fd = e->fd;
    pfd->revents = 0;
    poll_data.event_array[fd_count] = e;
    fd_count++;
    // fill in the event to request
    pfd->events = get_poll_event(e);
  }
  
  // do the poll
  poll_result = poll(poll_data.poll_array, fd_count, timeout_milliseconds);
  
  // add events to fire queue, note: O(n)
  if (poll_result > 0) {
    int i;
    int added_count = 0;
    for (i = 0; i < fd_count && added_count < poll_result; i++) {
      int revents = poll_data.poll_array[i].revents;
      eel_event_type event_type = poll_data.event_array[i]->type;
      // add all those that match
      if (// check for read availability
           (((revents & POLLIN) || (revents & POLLPRI)) 
             && (event_type == EEL_READ_EVENT)) || 
          // check for write availability
           ((revents & POLLOUT) 
             && (event_type == EEL_WRITE_EVENT)) || 
          // check for error occurrence
           (((revents & POLLERR) || (revents & POLLHUP) || (revents & POLLNVAL)) 
             && (event_type == EEL_ERROR_EVENT)) )
      {
        eel->add_to_fire(poll_data.event_array[i]);
        added_count++;
      }
    }
  }
  
  return poll_result;
}
